class HtbException(Exception):
    """Base exception class for `hackthebox`"""
    pass


class ApiError(HtbException):
    """The API responded in an unexpected way"""


class AuthenticationException(HtbException):
    """An error authenticating to the API"""
    pass


class NotFoundException(HtbException):
    """The API returned a 404 response for this request"""
    pass


class MissingEmailException(AuthenticationException):
    """An email was not given where it was required"""
    pass


class MissingPasswordException(AuthenticationException):
    """A password was not given where it was required"""
    pass


class MissingOTPException(AuthenticationException):
    """An OTP was not given but 2FA is enabled"""
    pass


class IncorrectOTPException(AuthenticationException):
    """An OTP was given but not accepted"""
    pass


class VpnException(HtbException):
    """An error associated with the VPN"""
    pass


class CannotSwitchWithActive(VpnException):
    """Failed to switch VPN because the user has an active machine"""
    pass


class MachineException(HtbException):
    """An error associated with a machine"""
    pass


class TooManyResetAttempts(MachineException):
    """Error for too many reset attempts"""
    pass


class UnknownSolveException(HtbException):
    """An unknown solve type was passed"""
    pass


class SolveError(HtbException):
    """Exceptions for solving"""


class IncorrectFlagException(SolveError):
    """An incorrect flag was submitted"""
    pass

class UserAlreadySubmitted(SolveError):
    """Player has already completed the user flag for this box"""
    pass


class RootAlreadySubmitted(SolveError):
    """Player has already completed the root flag for this box"""


class IncorrectArgumentException(HtbException):
    """An incorrectly formatted argument was passed"""

    reason: str

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"IncorrectArgumentException(reason='{self.reason}')"

    def __init__(self, reason: str):
        self.reason = reason
    pass


class NoDockerException(HtbException):
    """A challenge was 'started' when no Docker is available"""
    pass


class NoDownloadException(HtbException):
    """A challenge was 'downloaded' when no download is available"""
    pass


class RateLimitException(HtbException):
    """An internal ratelimit to prevent spam was violated"""
    def __init__(self, message):
        print(message)
        super().__init__(message)


class CacheException(HtbException):
    """There was an issue with the token cache"""
    pass



from typing import List, Iterator

from . import htb
from .team import Team
from .user import User


class Leaderboard(htb.HTBObject):
    """ The class representing a Leaderboard

    Args:
        data: A list of Leaderboard entries
        leaderboard_type: The Type of entries in the Leaderboard

    """

    _type: type
    _items: List[htb.HTBObject]
    _iter: Iterator[None]

    def __getitem__(self, key):
        return self._items[key]

    def __iter__(self):
        # noinspection PyTypeChecker
        self._iter = iter(self._items)
        return self

    def __next__(self):
        return next(self._iter)

    def __len__(self):
        return len(self._items)

    def __init__(self, data: List[dict], client: htb.HTBClient, leaderboard_type: type):
        self._type = leaderboard_type
        self._client = client
        if leaderboard_type == User:
            self._items = [User(usr, client, summary=True) for usr in data]
        elif leaderboard_type == Team:
            self._items = [Team(team, client, summary=True) for team in data]
        elif leaderboard_type == Country:
            self._items = [Country(country) for country in data]
        elif leaderboard_type == University:
            self._items = [University(university) for university in data]


class Country(htb.HTBObject):
    """ The class representing a Country

    Attributes:
        rank: The Country's global rank
        country_code: The Country's country code
        members: The number of members from the Country
        points: The Country's total points
        user_owns: The Country's total user owns
        root_owns: The Country's total root owns
        challenge_owns: The Country's total challenge owns
        user_bloods: The Country's total user bloods
        root_bloods: The Country's total root bloods
        fortress: The Country's total Fortress flags
        endgame: The Country's total Endgame flags
        name: The name of the Country
    Args:
        data: The data of the country

    """

    # TODO: Move this into its own file, maybe `misc.py`?

    rank: int
    country_code: str
    members: int
    points: int
    user_owns: int
    root_owns: int
    challenge_owns: int
    user_bloods: int
    root_bloods: int
    fortress: int
    endgame: int
    name: str

    def __init__(self, data: dict):
        self.rank = data['rank']
        self.country_code = data['country']
        self.members = data['members']
        self.points = data['points']
        self.user_owns = data['user_owns']
        self.root_owns = data['root_owns']
        self.challenge_owns = data['challenge_owns']
        self.user_bloods = data['user_bloods']
        self.root_owns = data['root_bloods']
        self.fortress = data['fortress']
        self.endgame = data['endgame']
        self.name = data['name']


class University(htb.HTBObject):
    """ The class representing a University

    Attributes:
        rank: The University's global rank
        students: The number of students from the University
        points: The University's total points
        user_owns: The University's total user owns
        root_owns: The University's total root owns
        challenge_owns: The University's total challenge owns
        user_bloods: The University's total user bloods
        root_bloods: The University's total root bloods
        fortress: The University's total Fortress flags
        endgame: The University's total Endgame flags
        name: The name of the University
    Args:
        data: The data of the University

    """

    rank: int
    students: int
    points: int
    user_owns: int
    root_owns: int
    challenge_owns: int
    user_bloods: int
    root_bloods: int
    fortress: int
    endgame: int
    name: str

    def __init__(self, data: dict):
        self.rank = data['rank']
        self.students = data['students']
        self.points = data['points']
        self.user_owns = data['user_owns']
        self.root_owns = data['root_owns']
        self.challenge_owns = data['challenge_owns']
        self.user_bloods = data['user_bloods']
        self.root_owns = data['root_bloods']
        self.fortress = data['fortress']
        self.endgame = data['endgame']
        self.name = data['name']



"""
Examples:
    Retrieving the current VPN server::

        print(client.get_current_vpn_server())

    Switching to a given VPN server::

        req = input("What server? ")
        server = next(filter(lambda x: x.friendly_name == req), client.get_all_vpn_servers()))
        server.switch()
        server.download(path="/tmp/out.ovpn")

"""

from __future__ import annotations

import os

from . import htb
from .errors import VpnException, CannotSwitchWithActive

from typing import TYPE_CHECKING, cast
if TYPE_CHECKING:
    from .htb import HTBClient


class VPNServer(htb.HTBObject):
    """Class representing individual VPN servers provided by Hack The Box

    Attributes:
        friendly_name: Friendly name of the server

            Example: ``'US Free 1'``

        current_clients: The number of currently connected clients
        location: The physical location of the server

            Example: ``'US'``

    """

    friendly_name: str
    current_clients: int
    location: str
    _detailed_func = lambda x: None

    # noinspection PyUnresolvedReferences
    def __init__(self, data: dict, client: "HTBClient", summary=False):
        self._client = client
        self.id = data["id"]
        self.friendly_name = data["friendly_name"]
        self.current_clients = data["current_clients"]
        self.location = data["location"]
        self.summary = summary

    def __repr__(self):
        return f"<VPN Server '{self.friendly_name}'>"

    def __str__(self):
        return f"{self.friendly_name}"

    def switch(self) -> bool:
        """
        Switches the client to use this VPN server

        Returns: Whether the switch was completed successfully
        """
        # TODO: Throw exception on failure
        result = cast(dict, self._client.do_request(f"connections/servers/switch/{self.id}", post=True))
        if result["status"] is True:
            return True
        if result["message"] == "You must stop your active machine before switching VPN":
            raise CannotSwitchWithActive
        raise VpnException

    def download(self, path=None, tcp=False) -> str:
        """

        Args:
            path: The name of the OVPN file to download to. If none is provided, it is saved to the current directory.
            tcp: Download TCP instead of UDP

        Returns: The path of the file

        """
        if path is None:
            path = os.path.join(os.getcwd(), f"{self.friendly_name}.ovpn")
        url = f"access/ovpnfile/{self.id}/0"
        if tcp:
            # Funky URL
            url += "/1"
        data = self._client.do_request(url, download=True)
        # We can't download VPN packs for servers we're not assigned to
        if b'You are not assigned' in data:
            self.switch()
        data = cast(bytes, self._client.do_request(url, download=True))
        with open(path, 'wb') as f:
            f.write(data)
        return path



from . import htb

from typing import TYPE_CHECKING, cast, Optional
if TYPE_CHECKING:
    from .user import User


class Team(htb.HTBObject):
    """ The class representing Hack The Box teams

    Attributes:
        name: The name of the Team
        points: The Team's total points
        motto: The Team motto
        description: The Team description
        country_name: The name of the country the Team is based in
        avatar_url: The relative URL of the Tean's avatar
        twitter: The Team's Twitter account
        facebook: The Team's Facebook account
        discord: The Team's Discord
        public: Whether the Team is publicly visible
        can_delete_avatar: Whether the active User can delete the avatar
        is_respected: Whether the active User has respected the Team
        join_request_sent: Whether the active User has sent a request to join the Team

    """
    name: str

    _detailed_attributes = ('points', 'motto', 'description', 'country_name', 'avatar_url', 'cover_image_url',
                            'twitter', 'facebook', 'discord', 'public', 'can_delete_avatar', 'captain',
                            'is_respected', 'join_request_sent')
    points: int
    motto: str
    description: str
    country_name: str
    avatar_url: str
    cover_image_url: str
    twitter: str
    facebook: str
    discord: str
    public: bool
    can_delete_avatar: bool
    # noinspection PyUnresolvedReferences
    _captain: Optional["User"] = None
    is_respected: Optional[bool] = None
    join_request_sent: Optional[bool] = None
    _ranking: Optional[int] = None
    _captain_id: int

    def __repr__(self):
        return f"<Team '{self.name}'>"

    def __init__(self, data: dict, client: htb.HTBClient, summary: bool = False):
        self._client = client
        self._detailed_func = client.get_team  # type: ignore
        self.id = data['id']
        self.name = data['name']
        if not summary:
            self.points = data['points']
            self.motto = data['motto']
            self.description = data['description']
            self.country_name = data['country_name']
            self.avatar_url = data['avatar_url']
            self.cover_image_url = data['cover_image_url']
            self.twitter = data['twitter']
            self.facebook = data['facebook']
            self.discord = data['facebook']
            self.public = data['public']
            self.can_delete_avatar = data['can_delete_avatar']
            self._captain_id = data['captain']['id']
            self.is_respected = data['is_respected']
            self.join_request_sent = data['join_request_sent']
        else:
            self._is_summary = True

    @property
    def ranking(self) -> int:
        """Retrieve the global ranking of the team

        Returns:

        """
        if not self._ranking:
            data = cast(dict, self._client.do_request(f"team/stats/owns/{self.id}"))
            self._ranking = data['rank']
        return cast(int, self._ranking)

    # noinspection PyUnresolvedReferences
    @property
    def captain(self) -> "User":
        from .user import User
        if not self._captain:
            self._captain = self._client.get_user(self._captain_id)
        return cast(User, self._captain)



from __future__ import annotations

import atexit
import base64
import getpass
import json
import os
import time
from typing import List, Callable, Union, Optional, Tuple, cast, Any, TYPE_CHECKING

import requests

from .constants import API_BASE, USER_AGENT
from .errors import AuthenticationException, NotFoundException, \
    IncorrectOTPException, ApiError

if TYPE_CHECKING:
    from .user import User
    from .search import Search
    from .machine import Machine, MachineInstance
    from .challenge import Challenge
    from .endgame import Endgame
    from .fortress import Fortress
    from .team import Team
    from .leaderboard import Leaderboard
    from .vpn import VPNServer


def jwt_expired(token: str) -> bool:
    """ Checks if a JWT token is expired

    Args:
        token: A JWT string - 3 Base64 sequences joined with .

    Returns:
        If the token is expired

    """
    payload = base64.b64decode(token.split('.')[1] + "==").decode()
    if time.time() > json.loads(payload)['exp']:
        return True
    else:
        return False


class HTBClient:
    """The client via which API requests are made

    Examples:
        Connecting to the API::

            from hackthebox import HTBClient
            client = HTBClient(email="user@example.com", password="S3cr3tP455w0rd!")
    Attributes:
        challenge_cooldown: Time when next download is allowed

    """
    # noinspection PyUnresolvedReferences
    _user: Optional["User"] = None
    _access_token: Optional[str]
    _refresh_token: Optional[str]
    _app_token: Optional[str]
    _api_base: str
    challenge_cooldown: int = 0

    def _refresh_access_token(self):
        """

        Use a saved refresh token to gain a new access token
        when the current one expires

        """
        headers = {"User-Agent": USER_AGENT}
        r = requests.post(self._api_base + "login/refresh", json={
            "refresh_token": self._refresh_token
        }, headers=headers)
        data = r.json()['message']
        if isinstance(data, str) and data.startswith("Unauthenticated"):
            raise AuthenticationException
        self._access_token = data['access_token']
        self._refresh_token = data['refresh_token']

    def do_request(self, endpoint, json_data=None, data=None, authorized=True, download=False, post=False) -> Union[
            dict, bytes]:
        """

        Args:
            endpoint: The API endpoint to request
            json_data: Data to be sent in JSON format
            data: Data to be sent in application/x-www-form-urlencoded format
            authorized: If the request requires an Authorization header
            download: If we are downloading raw data
            post: Force POST request
        Returns:
            The JSON response from the API or the raw data (if `download` is set)

        """
        headers = {"User-Agent": USER_AGENT}
        if authorized:
            # Don't use authorization if the API base URL isn't the real one -
            # i.e. we're running a test
            if self._app_token is not None:
                headers['Authorization'] = "Bearer " + self._app_token
            elif self._access_token is not None and self._refresh_token is not None:
                if jwt_expired(self._access_token):
                    self._refresh_access_token()
                headers['Authorization'] = "Bearer " + self._access_token
            else:
                raise AuthenticationException("No authentication tokens available")
        while True:
            if not json_data and not data:
                if post:
                    r = requests.post(self._api_base + endpoint, headers=headers, stream=download)
                else:
                    r = requests.get(self._api_base + endpoint, headers=headers, stream=download)
            else:
                r = requests.post(
                    self._api_base + endpoint, json=json_data, data=data, headers=headers, stream=download
                )
            if r.status_code != 429:
                break
            # Not sure on the exact ratelimit - loop until we don't get 429
            else:
                time.sleep(1)
        if r.status_code == 404:
            raise NotFoundException
        if download:
            return r.content
        else:
            return r.json()

    def __init__(self, email: Optional[str] = None, password: Optional[str] = None, otp: Optional[str | int] = None,
                 cache: Optional[str] = None, api_base: str = API_BASE, remember: Optional[bool] = False,
                 app_token: Optional[str] = None):
        """
        Authenticates to the API.

        If `cache` is set, the client will attempt to load access tokens from the given path. If they cannot be found,
        or are expired, normal API authentication will take place, and the tokens will be dumped to the file for the
        next launch.

        Args:
            email: The authenticating user's email address
            password: The authenticating user's password
            otp: The current OTP of the user, if 2FA is enabled
            cache: The path to load/store access tokens from/to
            remember: Whether to create a long-lasting 'remember me' token
            app_token: Authenticate using a provided App Token
        """
        self._api_base = api_base
        if cache is not None:
            if self.load_from_cache(cache) is False:
                self.do_login(email, password, otp, remember, app_token)
                self.dump_to_cache(cache)
            # Make sure we dump our current tokens out when we exit
            atexit.register(self.dump_to_cache, cache)
        else:
            self.do_login(email, password, otp, remember, app_token)

    def load_from_cache(self, cache: str) -> bool:
        """
        Args:
            cache: The cache file path

        Returns: Whether loading from the cache was successful
        """
        if not os.path.exists(cache):
            return False
        with open(cache, 'r') as f:
            data = json.load(f)
        self._access_token = data.get('access_token')
        self._refresh_token = data.get('refresh_token')
        self._app_token = data.get('app_token')
        if self._access_token is not None:
            if jwt_expired(self._access_token):
                try:
                    self._refresh_access_token()
                # Our refresh token is also invalid, we must log in again
                except AuthenticationException:
                    return False
        return True

    def dump_to_cache(self, cache):
        """
        Dumps the current access and refresh tokens to a file
        Args:
            cache: The path to the cache file
        """
        with open(cache, 'w') as f:
            json.dump({
                "access_token": self._access_token,
                "refresh_token": self._refresh_token,
                "app_token": self._app_token
            }, f)

    def do_login(self, email: Optional[str] = None, password: Optional[str] = None, otp: Optional[str | int] = None,
                 remember: Optional[bool] = False, app_token: Optional[str] = None):
        """
        Authenticates against the API. If credentials are not provided, they will be prompted for.
        """

        self._app_token = app_token
        if app_token is not None:
            self._access_token = self._refresh_token = None
        else:
            if email is None:
                email = input("Email: ")
            if password is None:
                password = getpass.getpass()

            data = cast(dict, self.do_request("login", json_data={
                "email": email, "password": password, "remember": remember
            }, authorized=False))
            msg = data['message']

            self._access_token = msg.get('access_token')
            if self._access_token is None:
                raise ApiError(f"Failed to get access token: {msg}")
            self._refresh_token = msg.get('refresh_token')
            if self._refresh_token is None:
                raise ApiError(f"Failed to get refresh token: {msg}")
            if data['message']['is2FAEnabled'] is True:
                if otp is None:
                    otp = input("OTP: ")
                if type(otp) == int:
                    # Optimistically try and create a string
                    otp = f"{otp:06d}"
                resp = cast(dict, self.do_request("2fa/login", json_data={
                    "one_time_password": otp
                }))
                if "correct" not in resp['message']:
                    raise IncorrectOTPException

    # noinspection PyUnresolvedReferences
    def search(self, search_term: str) -> "Search":
        """

        Args:
            search_term: The search term to pass to the API

        Returns: A Search object with lists of the items retrieved from the API

        """
        from .search import Search
        return Search(search_term, self)

    # noinspection PyUnresolvedReferences
    def get_machine(self, machine_id: int | str) -> "Machine":
        """

        Args:
            machine_id: The platform ID or name of the `Machine` to fetch

        Returns: The requested `Machine`

        """
        from .machine import Machine
        data = cast(dict, self.do_request(f"machine/profile/{machine_id}"))['info']
        return Machine(data, self)

    # noinspection PyUnresolvedReferences
    def get_todo_machines(self, limit: int = None) -> List[int]:
        """

        Retrieve a list of `Machine` ID's from the API based on the users todo list

        Args:
            limit: The maximum number to fetch

        Returns: A list of `Machine` ID's

        """
        data = cast(dict, self.do_request("home/user/todo"))['data']['machines'][:limit]
        return [m['id'] for m in data]


    # noinspection PyUnresolvedReferences
    def get_active_machine(self, release_arena: bool=False) -> Optional["MachineInstance"]:
        """

        Retrieve `Machine` currently assigned to user

        Returns: The `Machine` currently assigned (or active) to user

        """
        from .machine import Machine, MachineInstance
        if release_arena:
            info = cast(dict, self.do_request(f"release_arena/active"))['info']
        else:
            info = cast(dict, self.do_request(f"machine/active"))['info']
        if info:
            box = self.get_machine(info['id'])
            server = box._client.get_current_vpn_server(release_arena)
            return MachineInstance(box.ip, server, box, box._client)
        return None
        

    # noinspection PyUnresolvedReferences
    def get_machines(self, limit: int = None, retired: bool = False) -> List["Machine"]:
        """

        Retrieve a list of `Machine` from the API

        Args:
            limit: The maximum number to fetch
            retired: Whether to fetch from the retired list instead of the active list

        Returns: A list of `Machine`

        """
        from .machine import Machine
        if not retired:
            data = cast(dict, self.do_request("machine/list"))['info'][:limit]
        else:
            data = cast(dict, self.do_request("machine/list/retired"))['info'][:limit]
        machines = [Machine(m, self, summary=True) for m in data]
        for machine in machines:
            machine.retired = retired
        return machines

    # noinspection PyUnresolvedReferences
    def get_challenge(self, challenge_id: int | str) -> "Challenge":
        """

        Args:
            challenge_id: The platform ID or name of the `Challenge` to fetch

        Returns: The requested `Challenge`

        """
        from .challenge import Challenge
        data = cast(dict, self.do_request(f"challenge/info/{challenge_id}"))['challenge']
        return Challenge(data, self)

    # noinspection PyUnresolvedReferences
    def get_challenges(self, limit=None, retired=False) -> List["Challenge"]:
        """Requests a list of `Challenge` from the API

        Args:
            limit: The maximum number of `Challenge` to fetch
            retired: Whether to fetch from the retired list instead of the active list

        Returns: A list of `Challenge`

        """
        from .challenge import Challenge
        if retired:
            data = cast(dict, self.do_request("challenge/list/retired"))
        else:
            data = cast(dict, self.do_request("challenge/list"))
        challenges = []
        for challenge in data['challenges'][:limit]:
            challenges.append(Challenge(challenge, self, summary=True))
        return challenges

    # noinspection PyUnresolvedReferences
    def get_endgame(self, endgame_id: int) -> "Endgame":
        """Requests an Endgame from the API

        Args:
            endgame_id: The ID of the Endgame to fetch

        Returns: An Endgame

        """
        from .endgame import Endgame
        data = cast(dict, self.do_request(f"endgame/{endgame_id}"))["data"]
        return Endgame(data, self)

    # noinspection PyUnresolvedReferences
    def get_endgames(self, limit: int = None) -> List["Endgame"]:
        """Requests a list of Endgames from the API

        Args:
            limit: The maximum number of Endgames to fetch

        Returns: A list of Endgames

        """
        from .endgame import Endgame
        data = cast(dict, self.do_request(f"endgames"))["data"][:limit]
        endgames = []
        for endgame in data:
            endgames.append(Endgame(endgame, self, summary=True))
        return endgames

    # noinspection PyUnresolvedReferences
    def get_fortress(self, fortress_id: int) -> "Fortress":
        """Requests an Fortress from the API

        Args:
            fortress_id: The ID of the Fortress to fetch

        Returns: A Fortresse

        """
        from .fortress import Fortress
        data = cast(dict, self.do_request(f"fortress/{fortress_id}"))["data"]
        return Fortress(data, self)

    # noinspection PyUnresolvedReferences
    def get_fortresses(self, limit: int = None) -> List["Fortress"]:
        """Requests a list of Fortresses from the API

        Args:
            limit: The maximum number of Fortresses to fetch

        Returns: A list of Fortresses

        """
        from .fortress import Fortress
        data = cast(dict, self.do_request(f"fortresses"))["data"]
        fortresses = []
        # For some  reason, the fortress list is in the format {"1": <fortress1>, "2": <fortress2>}
        # instead of [<fortress1>, <fortress2>], meaning we have to sort it ourselves
        for fortress_id, fortress in sorted(data.items())[:limit]:
            fortresses.append(Fortress(fortress, self, summary=True))
        return fortresses

    # noinspection PyUnresolvedReferences
    def get_user(self, user_id: int) -> "User":
        """

        Args:
            user_id: The platform ID of the `User` to fetch

        Returns: The requested `User`

        """
        from .user import User
        data = cast(dict, self.do_request(f"user/profile/basic/{user_id}"))['profile']
        return User(data, self)

    # noinspection PyUnresolvedReferences
    def get_team(self, team_id: int) -> "Team":
        """

        Args:
            team_id: The platform ID of the `Team` to fetch

        Returns: The requested `Team`

        """
        from .team import Team
        data = cast(dict, self.do_request(f"team/info/{team_id}"))
        return Team(data, self)

    # noinspection PyUnresolvedReferences
    def get_hof(self, vip: bool = False) -> "Leaderboard":
        """
        Returns: A Leaderboard of the top 100 Users
        """
        from .leaderboard import Leaderboard
        from .user import User
        endpoint = "rankings/users"
        if vip:
            endpoint += "?vip=1"
        data = cast(dict, self.do_request(endpoint))['data']
        return Leaderboard(data, self, User)

    # noinspection PyUnresolvedReferences
    def get_hof_countries(self) -> "Leaderboard":
        """
        Returns: A Leaderboard of the top 100 Countries
        """
        from .leaderboard import Leaderboard, Country
        data = cast(dict, self.do_request("rankings/countries"))['data']
        return Leaderboard(data, self, Country)

    # noinspection PyUnresolvedReferences
    def get_hof_teams(self) -> "Leaderboard":
        """

        Returns: A Leaderboard of Teams

        """
        from .leaderboard import Leaderboard
        from .team import Team
        data = cast(dict, self.do_request("rankings/teams"))['data']
        return Leaderboard(data, self, Team)

    # noinspection PyUnresolvedReferences
    def get_hof_universities(self) -> "Leaderboard":
        """
        Returns: A Leaderboard of Universities

        """
        from .leaderboard import Leaderboard, University
        data = cast(dict, self.do_request("rankings/universities"))['data']
        return Leaderboard(data, self, University)

    # noinspection PyUnresolvedReferences
    def get_current_vpn_server(self, release_arena=False) -> VPNServer:
        """
        Returns: The currently assigned `VPNServer`

        Args:
            release_arena: Get the current release arena VPN server
        """
        from .vpn import VPNServer
        if release_arena:
            connections = cast(dict, self.do_request('connections/servers?product=release_arena'))['data']
            data = connections['assigned']
        else:
            connections = cast(dict, self.do_request('connections'))['data']
            data = connections['lab']['assigned_server']

        return VPNServer(data, self)

    # noinspection PyUnresolvedReferences
    def get_all_vpn_servers(self, release_arena=False) -> "List[VPNServer]":
        """
        Returns: A list of `VPNServer`

        Args:
            release_arena: Use the release arena VPN servers
        """
        from .vpn import VPNServer
        if release_arena:
            data = cast(dict, self.do_request("connections/servers?product=release_arena"))["data"]["options"]
        else:
            data = cast(dict, self.do_request("connections/servers?product=labs"))["data"]["options"]
        servers = []
        for location in data.keys():  # 'EU'
            for location_role in data[location].keys():  # 'EU - Free'
                for server in data[location][location_role]["servers"].values():
                    servers.append(VPNServer(server, self))
        return servers

    # noinspection PyUnresolvedReferences
    @property
    def user(self) -> "User":
        """

        Returns: The `User` associated with the current `HTBClient`

        """
        if not self._user:
            uid = cast(dict, self.do_request("user/info"))['info']['id']
            self._user = self.get_user(uid)
        return self._user


class HTBObject:
    """ Base class of all API objects

    Attributes:
        id: The ID of the associated object
    """
    _client: HTBClient
    # Attributes not fetched by a summary
    _detailed_attributes: Tuple[str, ...]
    _detailed_func: Callable[..., Any]
    _is_summary: bool = False
    id: int

    def __getattr__(self, item):
        """Retrieve attributes not given when initialised from a summary

        Some endpoints only provide a subset of the attributes available for a given object.
        If these extra attributes are requested, the object will request the full data from the
        API and fill out the missing items.

        Args:
            item: The name of the property to retrieve

        """
        if item in self._detailed_attributes and self._is_summary:
            new_obj = self._detailed_func(self.id)
            for attr in self._detailed_attributes:
                setattr(self, attr, getattr(new_obj, attr))
            self._is_summary = False
            return getattr(new_obj, item)
        else:
            raise AttributeError

    def __eq__(self, other):
        return self.id == other.id and type(self) == type(other)



from typing import List, cast, Optional

from .user import User
from .machine import Machine
from .team import Team
from .challenge import Challenge
from .errors import NotFoundException
from . import htb


class Search:
    """ Representation of a search on the platform

    Attributes:
        users: Users returned by the Search
        machines: Machines returned by the Search
        teams: Teams returned by the Search
        challenges: Challenges returned by the Search
        items: A dict of all items returned by the Search

    Args:
        search: The term to search for
        _tags: The list of tags to filter by
    """

    _users: Optional[List[User]] = None
    _machines: Optional[List[Machine]] = None
    _teams: Optional[List[Team]] = None
    _challenges: Optional[List[Challenge]] = None

    _user_ids: List[int]
    _machine_ids: List[int]
    _team_ids: List[int]
    _challenge_ids: List[int]

    _is_resolved: bool = False
    _term: str

    # The search API can return non-existent items (i.e. deleted). This should be handled and
    # not passed back to the user.

    @property
    def users(self) -> List[User]:
        if self._users is None:
            self._users = []
            for uid in self._user_ids:
                try:
                    self._users.append(self._client.get_user(uid))
                except NotFoundException:
                    pass
        return self._users

    @property
    def machines(self) -> List[Machine]:
        if self._machines is None:
            self._machines = []
            for uid in self._machine_ids:
                try:
                    self._machines.append(self._client.get_machine(uid))
                except NotFoundException:
                    pass
        return self._machines

    @property
    def teams(self) -> List[Team]:
        if self._teams is None:
            self._teams = []
            try:
                for uid in self._team_ids:
                    self._teams.append(self._client.get_team(uid))
            except NotFoundException:
                pass
        return self._teams

    @property
    def challenges(self) -> List[Challenge]:
        if self._challenges is None:
            self._challenges = []
            for uid in self._challenge_ids:
                try:
                    self._challenges.append(self._client.get_challenge(uid))
                except NotFoundException:
                    pass
        return self._challenges

    @property
    def items(self) -> dict:
        self._is_resolved = True
        return {
            "users": self.users,
            "machines": self.machines,
            "teams": self.teams,
            "challenges": self.challenges
        }

    # noinspection PyStatementEffect
    def __len__(self):
        return len(self._user_ids) + len(self._team_ids) + len(self._machine_ids) + len(self._challenge_ids)

    def __repr__(self):
        return f"<Search '{self._term}': {len(self)} results>"

    def __str__(self):
        return repr(self)

    def __init__(self, search: str, client: htb.HTBClient, _tags=None):
        if _tags is None:
            _tags = []
        self._term = search
        self._client = client
        # Ignoring tags - they seem to currently not work on the API level
        search_data = cast(dict, self._client.do_request("search/fetch?query=" + search))
        self._user_ids = [x['id'] for x in search_data.get("users", [])]
        self._machine_ids = [x['id'] for x in search_data.get("machines", [])]
        self._team_ids = [x['id'] for x in search_data.get("teams", [])]
        self._challenge_ids = [x['id'] for x in search_data.get("challenges", [])]



from datetime import datetime, timedelta
from typing import List, Union, cast, Optional, TYPE_CHECKING

import dateutil.parser

from . import htb, vpn
from .errors import (
    IncorrectArgumentException,
    IncorrectFlagException,
    TooManyResetAttempts,
    MachineException,
    UserAlreadySubmitted,
    RootAlreadySubmitted,
    SolveError,
)
from .solve import MachineSolve
from .utils import parse_delta

if TYPE_CHECKING:
    from .user import User


class Machine(htb.HTBObject):
    """The class representing Hack The Box machines

    Attributes:
        name: The Machine name
        os: The name of the operating system
        points: The points awarded for completion
        release_date: The date the Machine was released
        user_owns: The number of user owns the Machine has
        root_owns: The number of root owns the Machine has
        free: Whether the Machine is available on free servers
        user_owned: Whether the active User has owned the Machine's user account
        root_owned: Whether the active User has owned the Machine's user account
        reviewed: Whether the active User has reviewed the Machine
        stars: The average star rating of the Machine
        avatar: The relative URL of the Machine avatar
        difficulty: The difficulty of the machine
        :noindex: ip: The IP address of the machine

        active: Whether the Machine is active
        retired: Whether the Machine is retired
        avg_difficulty: The average numeric difficulty of the Machine
        completed: Whether the active User has completed the Machine
         :noindex: user_own_time: How long the active User took to own user
         :noindex: root_own_time: How long the active User took to own root
        user_blood: The Solve of the Machine's first user blood
        root_blood: The Solve of the Machine's first root blood
        user_own_time: How long the first User took to own user
        root_own_time: How long the first User took to own root
        difficulty_ratings: A dict of difficulty ratings given
    """

    name: str
    os: str
    points: int
    release_date: datetime
    root_owns: int
    free: bool
    user_owned: bool
    root_owned: bool
    reviewed: bool
    stars: float
    avatar: str
    difficulty: str

    _detailed_attributes = (
        "active",
        "retired",
        "user_own_time",
        "root_own_time",
        "user_blood",
        "root_blood",
        "user_blood_time",
        "root_blood_time",
        "difficulty_ratings",
    )
    active: bool
    retired: bool
    avg_difficulty: int
    completed: bool
    user_own_time: timedelta
    root_own_time: timedelta
    user_blood: MachineSolve
    root_blood: MachineSolve
    user_blood_time: timedelta
    root_blood_time: timedelta
    difficulty_ratings: dict

    # noinspection PyUnresolvedReferences
    _authors: Optional[List["User"]] = None
    _author_ids: List[int]
    _is_release: Optional[bool] = None
    _ip: Optional[str] = None

    def submit(self, flag: str, difficulty: int) -> str:
        """Submits a flag for a Machine

        Args:
            flag: The flag for the Machine
            difficulty: A rating between 10 and 100 of the Machine difficulty.
                        Must be a multiple of 10.
        """
        if difficulty < 10 or difficulty > 100 or difficulty % 10 != 0:
            raise IncorrectArgumentException(
                reason="Difficulty must be a multiple of 10, between 10 and 100"
            )

        submission = cast(
            dict,
            self._client.do_request(
                "machine/own",
                json_data={"flag": flag, "id": self.id, "difficulty": difficulty},
            ),
        )
        if submission["status"] == 400:
            if submission["message"] == "Incorrect flag!":
                raise IncorrectFlagException
            elif submission["message"] == f"{self.name} user is already owned.":
                raise UserAlreadySubmitted
            elif submission["message"] == f"{self.name} root is already owned.":
                raise RootAlreadySubmitted
            else:
                raise SolveError
        return submission["message"]

    # noinspection PyUnresolvedReferences
    @property
    def authors(self) -> List["User"]:
        """Fetch the author(s) of the Machine

        Returns: List of Users

        """
        if not self._authors:
            self._authors = []
            for uid in self._author_ids:
                self._authors.append(self._client.get_user(uid))
        return self._authors

    @property
    def is_release(self):
        if self._is_release is not None:
            return self._is_release
        self._is_release = False

        data = self._client.do_request("connections")["data"]
        try:
            if data["release_arena"]["machine"]["id"] == self.id:
                self._is_release = True
        except AttributeError:
            pass
        return self._is_release

    @property
    def ip(self):
        """The IP of an active machine."""
        if self._ip is not None:
            return self._ip
        listing = self._client.do_request("machine/list")["info"]
        m = next(filter(lambda x: x["id"] == self.id, listing))
        self._ip = m["ip"]
        return self._ip

    def start(self, release_arena=False) -> Union["MachineInstance", None]:
        """Alias for `Machine.spawn()`"""
        return self.spawn(release_arena)

    def spawn(self, release_arena=False) -> "MachineInstance":
        """Spawn an instance of this machine.

        Args:
            release_arena: Whether to use Release Arena to spawn the machine

        Returns:
            The spawned `MachineInstance`
        """
        if release_arena:
            if not self.is_release:
                # TODO: Better exception
                raise Exception("Machine is not on release arena")
            data = cast(dict, self._client.do_request("release_arena/spawn", post=True))
            if data.get("success") != 1:
                raise Exception(f"Failed to spawn: {data}")
            ip = cast(dict, self._client.do_request("release_arena/active"))["info"][
                "ip"
            ]
            server = self._client.get_current_vpn_server(release_arena=True)
        else:
            data = cast(
                dict,
                self._client.do_request("vm/spawn", json_data={"machine_id": self.id}),
            )
            if "Machine deployed" in cast(
                str, data.get("message")
            ) or "You have been assigned" in cast(str, data.get("message")):
                ip = cast(dict, self._client.do_request(f"machine/profile/{self.id}"))[
                    "info"
                ]["ip"]
                server = self._client.get_current_vpn_server()
            else:
                raise Exception(f"Failed to spawn: {data}")
        return MachineInstance(ip, server, self, self._client)

    def __repr__(self):
        return f"<Machine '{self.name}'>"

    def __init__(self, data: dict, client: htb.HTBClient, summary: bool = False):
        self._client = client
        self._detailed_func = client.get_machine  # type: ignore
        self.id = data["id"]
        self.name = data["name"]
        self.os = data["os"]
        self.points = data["points"]
        self.release_date = dateutil.parser.parse(data["release"])
        self.user_owns = data["user_owns_count"]
        self.root_owns = data["root_owns_count"]
        self.user_owned = data["authUserInUserOwns"]
        self.root_owned = data["authUserInRootOwns"]
        self.reviewed = data["authUserHasReviewed"]
        self.stars = float(data["stars"])
        self.avatar = data["avatar"]
        self.difficulty = data["difficultyText"]
        self.free = data["free"]
        self._author_ids = [data["maker"]["id"]]
        if data.get("ip"):
            self._ip = data["ip"]
        if data["maker2"]:
            self._author_ids.append(data["maker2"]["id"])
        if not summary:
            self.active = bool(data["active"])
            self.retired = bool(data["retired"])
            if data["authUserInUserOwns"]:
                self.user_own_time = parse_delta(data["authUserFirstUserTime"])
            if data["authUserInRootOwns"]:
                self.root_own_time = parse_delta(data["authUserFirstRootTime"])
            self.difficulty_ratings = data["feedbackForChart"]
            if data["userBlood"]:
                user_blood_data = {
                    "date": dateutil.parser.parse(data["userBlood"]["created_at"]),
                    "first_blood": True,
                    "id": data["id"],
                    "name": data["name"],
                    "type": "user",
                }
                self.user_blood = MachineSolve(user_blood_data, self._client)
                self.user_blood_time = parse_delta(
                    data["userBlood"]["blood_difference"]
                )
            if data["rootBlood"]:
                user_blood_data = {
                    "date": dateutil.parser.parse(data["rootBlood"]["created_at"]),
                    "first_blood": True,
                    "id": data["id"],
                    "name": data["name"],
                    "type": "root",
                }
                self.root_blood = MachineSolve(user_blood_data, self._client)
                self.root_blood_time = parse_delta(
                    data["rootBlood"]["blood_difference"]
                )
        else:
            self._is_summary = True


class MachineInstance:
    """Representation of an active machine instance

    Attributes:
        ip: The IP the instance can be reached at
        server: The `VPNServer` that the machine is on
        machine: The `Machine` this is an instance of
        client: The passed-through API client
    """

    ip: str
    server: vpn.VPNServer
    client: htb.HTBClient
    machine: Machine

    def __init__(
        self, ip: str, server: vpn.VPNServer, machine: Machine, client: htb.HTBClient
    ):
        self.client = client
        self.ip = ip
        self.server = server
        self.machine = machine

    def __repr__(self):
        return f"<'{self.machine.name}'@{self.server.friendly_name} - {self.ip}>"

    def stop(self):
        """Request the instance be stopped."""
        if self.machine.is_release:
            self.client.do_request("release_arena/terminate", post=True)
        else:
            self.client.do_request(
                "vm/terminate", json_data={"machine_id": self.machine.id}
            )

        # Can't delete references to the object from here so we just have
        # to set everything to None and prevent further usage
        self.server = None
        self.ip = None
        self.client = None
        self.machine = None

    def reset(self):
        """Request the instance be reset."""
        if self.machine.is_release:
            resp = self.client.do_request(
                "release_arena/reset", json_data={"machine_id": self.machine.id}
            )
        else:
            resp = self.client.do_request(
                "vm/reset", json_data={"machine_id": self.machine.id}
            )
        # VM
        if resp["message"].endswith(" will be reset in 1 minute."):
            return True
        elif resp["message"] == "Too many reset machine attempts. Try again later!":
            raise TooManyResetAttempts
        # RA
        elif resp["success"]:
            return True
        elif resp["message"].startswith("You must wait"):
            raise TooManyResetAttempts
        raise MachineException



from typing import List, cast, Optional

from . import htb
from .challenge import Challenge
from .errors import NotFoundException
from .machine import Machine


class Content:
    """Representation of a content fetch for a user on the platform

    Attributes:
        machines: Machines authored by the given user
        challenges: Challenges authored by the given user
        writeups: Currently not implemented; Write-ups associated with the given user
        items: A dict of all content items

    Args:
        userid: The user id
    """

    _machines: Optional[List[Machine]] = None
    _challenges: Optional[List[Challenge]] = None

    _machine_ids: List[int]
    _challenge_ids: List[int]

    _is_resolved: bool = False
    _user_id: int

    @property
    def machines(self) -> List[Machine]:
        if self._machines is None:
            self._machines = []
            for uid in self._machine_ids:
                try:
                    self._machines.append(self._client.get_machine(uid))
                except NotFoundException:
                    pass
        return self._machines

    @property
    def challenges(self) -> List[Challenge]:
        if self._challenges is None:
            self._challenges = []
            for uid in self._challenge_ids:
                try:
                    self._challenges.append(self._client.get_challenge(uid))
                except NotFoundException:
                    pass
        return self._challenges

    @property
    def items(self) -> dict:
        self._is_resolved = True
        return {"machines": self.machines, "challenges": self.challenges}

    def __len__(self):
        return len(self._machine_ids) + len(self._challenge_ids)

    def __repr__(self):
        return f"<Content '{self._user_id}': {len(self._machine_ids)} machines; {len(self._challenge_ids)} challenges>"

    def __str__(self):
        return repr(self)

    def __init__(self, userid: int, client: htb.HTBClient):
        self._user_id = userid
        self._client = client
        data = cast(dict, self._client.do_request(f"user/profile/content/{userid}"))[
            "profile"
        ]["content"]
        self._machine_ids = [x["id"] for x in data.get("machines", [])]
        self._challenge_ids = [x["id"] for x in data.get("challenges", [])]



from typing import List, cast, Optional

from . import htb
from .user import User
from .errors import IncorrectFlagException


class Endgame(htb.HTBObject):
    """ The class representing Hack The Box endgames

    Attributes:
        name: The name of the Endgame
        avatar: The relative URL of the Endgame's avatar
        cover_image: The relative URL of the Endgame's cover image
        retired: Whether the Endgame is retired
        vip: Whether the Endgame requires VIP
        points: The points awarded by the Endgame
        completions: The number of players who have completed the Endgame
        reset_votes: The number of votes to reset the Endgame
        entry_points: IP addresses the Endgame can be contacted on
        description: The HTML description of the Endgame

    """

    name: str
    avatar: str
    cover_image: str
    retired: bool
    vip: bool

    _detailed_attributes = ('points', 'completions', 'reset_votes', 'entry_points', 'description')
    points: int
    completions: int
    reset_votes: int
    entry_points: List[str]
    description: str

    _authors: Optional[List[User]] = None
    _author_ids: Optional[List[int]] = None

    def submit(self, flag: str):
        """ Submits a flag for an Endgame

        Args:
            flag: The flag for the Endgame

        """
        submission = cast(dict, self._client.do_request(f"endgame/{self.id}/flag", json_data={
            "flag": flag,
        }))
        if submission['message'] == "Wrong flag":
            raise IncorrectFlagException
        return True

    @property
    def authors(self):
        """
        The creators of the Endgame
        Returns: A list of Users

        """
        if self._authors is None:
            self._authors = []
            for user_id in self._author_ids:
                self._authors.append(self._client.get_user(user_id))
        return self._authors

    def __repr__(self):
        return f"<Endgame '{self.name}'>"

    def __init__(self, data: dict, client: htb.HTBClient, summary=False):
        self._client = client
        self._detailed_func = client.get_endgame  # type: ignore
        self.id = data['id']
        self.name = data['name']
        self.avatar = data['avatar_url']
        self.cover_image = data['cover_image_url']
        self.retired = data['retired']
        self._author_ids = []
        self.vip = data['vip']
        for user in data['creators']:
            self._author_ids.append(user['id'])
        if not summary:
            self.points = int(data['points'])
            self.entry_points = data['entry_points']
            self.completions = data['players_completed']
            self.reset_votes = data['endgame_reset_votes']
            self.description = data['description']
        else:
            self._is_summary = True



import re
from datetime import timedelta


def parse_delta(time: str) -> timedelta:
    """ Generates a timedelta from a string

    Args:
        time: The delta as a string

    Returns:
        A datetime.timedelta object

    """
    regex = re.compile(r'((?P<years>\d+?)[Yy])? ?((?P<months>\d+?)M)? ?'
                       r'((?P<weeks>\d+?)[Ww])? ?'
                       r'((?P<days>\d+?)[Dd])? ?((?P<hours>\d+?)[Hh])? ?'
                       r'((?P<minutes>\d+?)m)? ?((?P<seconds>\d+?)[Ss])?'
                       )
    parts = regex.match(time)
    if not parts or parts.group() == '':
        raise ValueError
    parts_dict = parts.groupdict()
    time_params = {'years': 0, 'months': 0, 'weeks': 0,
                   'days': 0, 'hours': 0, 'minutes': 0,
                   'seconds': 0}
    for (name, param) in parts_dict.items():
        if param:
            time_params[name] = int(param)
    # Remove unsupported params for timedelta
    time_params['days'] += time_params['years'] * 365
    del time_params['years']
    time_params['days'] += time_params['months'] * 30
    del time_params['months']
    return timedelta(**time_params)



from .challenge import Challenge
from .endgame import Endgame
from .errors import *
from .fortress import Fortress
from .htb import HTBClient, HTBObject
from .machine import Machine,MachineInstance
from .search import Search
from .solve import *
from .team import Team
from .user import User
from .vpn import VPNServer



from typing import List, Optional, TYPE_CHECKING

from hackthebox.content import Content

from . import htb
from .solve import MachineSolve, ChallengeSolve, EndgameSolve, FortressSolve, Solve


if TYPE_CHECKING:
    from .htb import HTBClient
    from .team import Team


class User(htb.HTBObject):
    """The class representing Hack The Box Users

    Attributes:
        name: The username of the User
        avatar: The relative URL of the User's avatar
        ranking: The User's position on the Hall of Fame
        points: The User's current total points
        user_owns: The User's total Machine user owns
        root_owns: The User's total Machine root owns
        user_bloods: The User's total Machine user bloods
        root_bloods: The User's total Machine root bloods
        rank_name: The name of the User's current rank
        country_name: The name of the User's country
        team: The User's Team
        public: Whether the User's profile is publicly visible

        timezone: The User's timezone
        vip: Whether the User is VIP
        vip_plus: Whether the user is VIP+
        respects: The number of respects the User has
        university: The User's University
        university_name: The User's university's name
        description: The User's description
        github: The User's Github profile
        linkedin: The User's LinkedIn profile
        twitter: The User's Twitter account
        website: The User's website
        respected: Whether the active User respects the User
        followed: Whether the active User follows the User
        rank_id: The ID of the User's rank
        rank_progress: The User's progress to the next rank
        next_rank: The next rank the User will reach
        next_rank_points: The points required to reach the next rank
        rank_requirement: The ownership required for the current Rank

    """

    name: str
    avatar: str
    ranking: int
    points: int
    root_owns: int
    user_owns: int
    root_bloods: int
    user_bloods: int
    rank_name: str

    _detailed_attributes = (
        "timezone",
        "vip",
        "vip_plus",
        "respects",
        "university",
        "university_name",
        "description",
        "github",
        "linkedin",
        "twitter",
        "website",
        "respected",
        "followed",
        "rank_id",
        "rank_progress",
        "next_rank",
        "next_rank_points",
        "rank_ownership",
        "rank_requirement",
        "country_name",
        "team",
        "public",
    )
    timezone: str
    vip: bool
    vip_plus: bool
    respects: int
    # TODO: University object
    university = None
    university_name: str
    description: str
    github: str
    linkedin: str
    twitter: str
    website: str
    respected: bool
    followed: bool
    rank_id: int
    rank_progress: int
    next_rank: str
    next_rank_points: int
    rank_ownership: float
    rank_requirement: int
    country_name: str
    # noinspection PyUnresolvedReferences
    team: "Team"
    public: bool

    _activity: Optional[List[Solve]] = None

    @property
    def activity(self):
        if not self._activity:
            self._activity = []
            solve_list = (self._client.do_request(f"user/profile/activity/{self.id}"))[
                "profile"
            ]["activity"]
            for solve_item in solve_list:
                solve_type = solve_item["object_type"]
                if solve_type == "machine":
                    self._activity.append(MachineSolve(solve_item, self._client))
                elif solve_type == "challenge":
                    self._activity.append(ChallengeSolve(solve_item, self._client))
                elif solve_type == "endgame":
                    self._activity.append(EndgameSolve(solve_item, self._client))
                elif solve_type == "fortress":
                    self._activity.append(FortressSolve(solve_item, self._client))

        return self._activity

    def __repr__(self):
        return f"<User '{self.name}'>"

    # noinspection PyUnresolvedReferences
    def __init__(self, data: dict, client: "HTBClient", summary: bool = False):
        """Initialise a `User` using API data"""
        self._client = client
        self._detailed_func = client.get_user  # type: ignore
        self.id = data["id"]
        self.name = data["name"]
        self.user_owns = data["user_owns"]
        self.points = data["points"]

        if summary:
            self._is_summary = True
            self.ranking = data["rank"]
            self.root_owns = data["root_owns"]
            self.user_bloods = data.get("user_bloods_count") or 0
            self.root_bloods = data.get("root_bloods_count") or 0
            self.rank_name = data.get("rank_text") or ""
        else:
            self.ranking = data["ranking"]
            self.root_owns = data["system_owns"]
            self.user_bloods = data["user_bloods"]
            self.root_bloods = data["system_bloods"]
            self.rank_name = data["rank"]

            self.respects = data["respects"]
            self.university = data["university"]
            self.university_name = data["university_name"]
            self.description = data["description"]
            self.github = data["github"]
            self.linkedin = data["linkedin"]
            self.twitter = data["twitter"]
            self.website = data["website"]
            self.respected = data.get("isRespected", False)
            self.followed = data.get("isFollowed", False)
            self.rank_progress = data["current_rank_progress"]
            self.next_rank = data["next_rank"]
            self.next_rank_points = data["next_rank_points"]
            self.rank_ownership = float(data["rank_ownership"])
            self.rank_requirement = data["rank_requirement"]
            self.country_name = data["country_name"]
            self.team = data["team"]
            self.public = bool(data["public"])

    # noinspection PyUnresolvedReferences
    def get_content(self):
        return Content(self.id, self._client)

    # noinspection PyUnresolvedReferences
    def get_machines(self):
        return self.get_content().machines

    # noinspection PyUnresolvedReferences
    def get_challenges(self):
        return self.get_content().challenges



from __future__ import annotations
from typing import List, cast

from . import htb
from .errors import IncorrectFlagException


class Fortress(htb.HTBObject):
    """ The class representing Hack The Box fortresses

    Attributes:
        id: The ID of the Fortress
        name: The name of the Fortress
        image: The relative URL of the Fortress' image
        num_flags: The number of available flags

        reset_votes: The number of votes to reset the Fortress
        progress: The active user's progress through the Fortress, out of 100
        flags: The list of flags available
        company: The Fortress' associated Company
        ip: IP address the Fortress can be contacted on

    """

    name: str
    image: str
    num_flags: int

    _detailed_attributes = ('reset_votes', 'progress', 'flags', 'company', 'ip')
    reset_votes: int
    progress: int
    flags: List
    company: Company
    ip: str

    def submit(self, flag: str):
        """ Submits a flag for an Fortress

        Args:
            flag: The flag for the Fortress

        """
        submission = cast(dict, self._client.do_request(f"fortress/{self.id}/flag", json_data={
            "flag": flag,
        }))
        if submission['message'] == "Wrong flag":
            raise IncorrectFlagException
        return True

    def __repr__(self):
        return f"<Fortress '{self.name}'>"

    def __init__(self, data: dict, client: htb.HTBClient, summary=False):
        self._client = client
        self._detailed_func = client.get_fortress  # type: ignore
        self.id = data['id']
        self.name = data['name']
        self.image = data['image']
        if summary:
            self.num_flags = data['number_of_flags']
            self._is_summary = True
        else:
            self.num_flags = len(data['flags'])
            self.company = Company(data['company'])
            self.reset_votes = data['reset_votes']
            self.progress = data['progress_percent']
            self.flags = data['flags']
            self.ip = data['ip']


class Company:
    """ Representation of a company registered on Hack The Box

    Attributes:
        id: The Company ID
        name: The Company name
        description: The Company description
        url: The Company website
        image: The Company logo

    """
    id: int
    name: str
    description: str
    url: str
    image: str

    def __init__(self, data: dict):
        self.id = data['id']
        self.name = data['name']
        self.description = data['description']
        self.url = data['url']
        self.image = data['image']



from . import htb
from datetime import datetime

from typing import Optional, TYPE_CHECKING, cast
if TYPE_CHECKING:
    from .htb import HTBClient, HTBObject
    from .machine import Machine


class Solve:
    """Representation of completion of Hack The Box content

    Attributes:
        id: The ID of the solved item
        name: The name of the solved item
        date: The date of the solve
        blood: Whether the solve was a first blood
        points: The points awarded from the solve
        item (HTBObject): The solved item

    """
    _client: "HTBClient"
    _item: Optional["HTBObject"] = None   # The solved item
    id: int
    name: str
    date: datetime
    blood: bool
    points: int

    def __init__(self, data: dict, client: "HTBClient"):
        self._client = client
        self.date = data['date']
        self.blood = data['first_blood']
        self.id = data['id']
        self.name = data['name']


class MachineSolve(Solve):
    """Representation of solving a Machine"""
    type: str   # User/Root

    def __repr__(self):
        return f"<Solve {self.type}@{self.name}>"

    @property
    def item(self):
        return self.machine

    # noinspection PyUnresolvedReferences
    @property
    def machine(self) -> "Machine":
        """The solved Machine"""
        from .machine import Machine
        if not self._item:
            self._item = self._client.get_machine(self.id)
        return cast(Machine, self._item)

    def __init__(self, data: dict, client: "htb.HTBClient"):
        super().__init__(data, client)
        self.type = data['type']


class ChallengeSolve(Solve):
    """Representation of solving a Challenge"""
    category: str

    def __repr__(self):
        return f"<Solve {self.name}@{self.category}>"

    @property
    def item(self):
        return self.challenge

    @property
    def challenge(self):
        """The solved Challenge"""
        if not self._item:
            self._item = self._client.get_challenge(self.id)
        return self._item

    def __init__(self, data: dict, client: "htb.HTBClient"):
        super().__init__(data, client)
        self.category = data['challenge_category']


class EndgameSolve(Solve):
    """Representation of solving a Endgame"""
    flag_name: str

    def __repr__(self):
        return f"<Solve {self.flag_name}@{self.name}>"

    @property
    def item(self):
        return self.endgame

    @property
    def endgame(self):
        """The solved Endgame"""
        if not self._item:
            self._item = self._client.get_endgame(self.id)
        return self._item

    def __init__(self, data: dict, client: "htb.HTBClient"):
        super().__init__(data, client)
        self.flag_name = data['flag_title']


class FortressSolve(Solve):
    """Representation of solving a Fortress"""
    flag_name: str

    def __repr__(self):
        return f"<Solve {self.flag_name}@{self.name}>"

    @property
    def item(self):
        return self.fortress

    @property
    def fortress(self):
        """The solved Fortress"""
        if not self._item:
            self._item = self._client.get_fortress(self.id)
        return self._item

    def __init__(self, data: dict, client: "htb.HTBClient"):
        super().__init__(data, client)
        self.flag_name = data['flag_title']



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="PyHackTheBox",
    version="0.5.6-1",
    author="clubby789@github.com",
    author_email="clubby789@gmail.com",
    description="A wrapper for the Hack The Box API.",
    long_description=long_description,
    url="https://github.com/clubby789/htb-api",
    project_urls={
        "Documentation": "https://pyhackthebox.readthedocs.io/en/latest/"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["hackthebox"],
    python_requires='>=3.8',
    install_requires=[
        "requests==2.27.1",
        "python-dateutil==2.8.2"
    ],
    long_description_content_type='text/markdown'
)



"""
Examples:
    Starting a challenge and submitting the flag::

        challenge = client.get_challenge(100)
        instance = challenge.start()
        r = remote(instance.ip, instance.port)
        # Do the challenge.....
        instance.stop()
        challenge.submit(flag, difficulty=50)

"""

from __future__ import annotations

import os
import time
from datetime import datetime
from typing import List, Optional, cast, TYPE_CHECKING

import dateutil.parser

from . import htb
from .constants import DOWNLOAD_COOLDOWN
from .errors import IncorrectFlagException, IncorrectArgumentException, NoDockerException, \
    NoDownloadException, RateLimitException

if TYPE_CHECKING:
    from .htb import HTBClient
    from .user import User


class Challenge(htb.HTBObject):
    """ The class representing Hack The Box challenges

    Attributes:
        name (str): The name of the challenge
        retired: Whether the challenge is retired
        difficulty: The official difficulty of the challenge
        avg_difficulty: The average user-given difficulty
        points: The points awarded on completion
        difficulty_ratings: A dict of difficulty ratings given
        solves: The number of solves a challenge has
        likes: The number of likes a challenge has
        dislikes: The number of dislikes a challenge has
        release_date: The date the challenge was released
        solved: Whether the active user has completed the challenge
        is_liked: Whether the active user has liked the challenge
        is_disliked: Whether the active user has disliked the challenge

        description: The challenge description
        category: The name of the category
        has_download: Whether the challenge has a download available
        has_docker: Whether the challenge has a remote instance available

    """
    name: str
    retired: bool
    difficulty: str
    avg_difficulty: int
    points: int
    difficulty_ratings = None
    solves: int
    likes: int
    dislikes: int
    release_date: datetime
    solved: bool
    is_liked: bool
    is_disliked: bool
    recommended: bool

    # noinspection PyUnresolvedReferences
    _authors: Optional[List["User"]] = None
    _author_ids: List[int]

    _detailed_attributes = ('description', 'category', 'has_download', 'has_docker', 'instance')
    description: str
    category: str
    has_download: bool
    has_docker: bool
    instance: Optional[DockerInstance]

    def submit(self, flag: str, difficulty: int):
        """ Submits a flag for a Challenge

        Args:
            flag: The flag for the Challenge
            difficulty: A rating between 10 and 100 of the Challenge difficulty.
                        Must be a multiple of 10.

        """
        if difficulty < 10 or difficulty > 100 or difficulty % 10 != 0:
            raise IncorrectArgumentException(reason="Difficulty must be a multiple of 10, between 10 and 100")

        submission = cast(dict, self._client.do_request("challenge/own", json_data={
            "flag": flag,
            "challenge_id": self.id,
            "difficulty": difficulty
        }))
        if submission['message'] == "Incorrect flag":
            raise IncorrectFlagException
        return True

    def start(self) -> DockerInstance:
        """
        Requests the challenge be started

        Returns:
            The DockerInstance that was started

        """
        if not self.has_docker:
            raise NoDockerException
        instance = cast(dict, self._client.do_request("challenge/start", json_data={"challenge_id": self.id}))
        # TODO: Handle failure to start
        self.instance = DockerInstance(instance['ip'], instance['port'], self.id, self._client, instance['id'])
        return self.instance

    def download(self, path=None) -> str:
        """

        Args:
            path: The name of the zipfile to download to. If none is provided, it is saved to the current directory.

        Returns: The path of the file

        """
        if not self.has_download:
            raise NoDownloadException
        if self._client.challenge_cooldown > time.time():
            raise RateLimitException("Challenge download ratelimit exceeded - please do not remove this")
        if path is None:
            path = os.path.join(os.getcwd(), f"{self.name}.zip")
        data = cast(bytes, self._client.do_request(f"challenge/download/{self.id}", download=True))
        self._client.challenge_cooldown = int(time.time()) + DOWNLOAD_COOLDOWN
        with open(path, 'wb') as f:
            f.write(data)
        return path

    # noinspection PyUnresolvedReferences
    @property
    def authors(self) -> List["User"]:
        """Fetch the author(s) of the Challenge

        Returns: List of Users

        """
        if not self._authors:
            self._authors = []
            for uid in self._author_ids:
                self._authors.append(self._client.get_user(uid))
        return self._authors

    def __repr__(self):
        return f"<Challenge '{self.name}'>"

    # noinspection PyUnresolvedReferences
    def __init__(self, data: dict, client: "HTBClient", summary: bool = False):
        """Initialise a `Challenge` using API data"""
        self._client = client
        self._detailed_func = client.get_challenge  # type: ignore
        self.id = data['id']
        self.name = data['name']
        self.retired = bool(data['retired'])
        self.points = int(data['points'])
        self.difficulty = data["difficulty"]
        self.difficulty_ratings = data['difficulty_chart']
        self.solves = data['solves']
        self.solved = data['authUserSolve']
        self.likes = data['likes']
        self.dislikes = data['dislikes']
        self.release_date = dateutil.parser.parse(data['release_date'])
        if not summary:
            self.description = data['description']
            self.category = data['category_name']
            self._author_ids = [data['creator_id']]
            if data['creator2_id']:
                self._author_ids.append(data['creator2_id'])
            self.has_download = data['download']
            self.has_docker = data['docker']
            if data['docker_ip']:
                self.instance = DockerInstance(data['docker_ip'], data['docker_port'], self.id, self._client)
            else:
                self.instance = None
        else:
            self._is_summary = True


class DockerInstance:
    """Representation of an active Docker container instance of a Challenge

    Attributes:
        container_id: The ID of the container
        port: The port the container is listening on
        ip: The IP the instance can be reached at
        chall_id: The connected challenge
        client: The passed-through API client

    """

    id: str
    port: int
    ip: str
    chall_id: int
    client: htb.HTBClient

    def __init__(self, ip: str, port: int,  chall_id: int, client: htb.HTBClient, container_id: str = None):
        self.client = client
        self.id = container_id or ""
        self.port = port
        self.ip = ip
        self.chall_id = chall_id

    def stop(self):
        """Request the instance be stopped. Zeroes out all properties"""
        self.client.do_request("challenge/stop", json_data={"challenge_id": self.chall_id})
        # TODO: Handle failures to stop
